# flake8: noqa

from pyproj import CRS
from pystac import Provider

USGS_3DEP_ID = "usgs-3dep"
THREEDEP_EPSG = 5498
THREEDEP_CRS = CRS.from_epsg(THREEDEP_EPSG)
PRODUCTS = ["1", "13"]
LICENSE = "PDDL-1.0"

DESCRIPTION = """The USGS 3D Elevation Program (3DEP) Datasets from The National Map are the primary elevation data product produced and distributed by the USGS. The 3DEP program provides a variety of resolution raster elevation data of the conterminous United States, Alaska, Hawaii, and the island territories. Some of the data sets such as the 1/3rd arc-second and 1 arc-second data set are derived from diverse source data sets that are processed to a specification with a consistent resolution, coordinate system, elevation units, and horizontal and vertical datums.  These seamless DEMs were referred to as the National Elevation Dataset (NED) from about 2000 through 2015 at which time they became the seamless DEM layers under the 3DEP program and the NED name and system were retired. Other 3DEP products include one-meter DEMs produced exclusively from high resolution light detection and ranging (lidar) source data and five-meter DEMs in Alaska as well as various source datasets including the lidar point cloud and interferometric synthetic aperture radar (Ifsar) digital surface models and intensity images. All 3DEP products are public domain. The 3DEP program is the logical result of the maturation of the long-standing USGS elevation program, which for many years concentrated on production of topographic map quadrangle-based digital elevation models. The 3DEP data  serves as the elevation layer of The National Map, and provides basic elevation information for earth science studies and mapping applications in the United States.

The seamless DEM layers under the 3DEP program are a multi-resolution dataset that is updated continuously to integrate newly available, improved elevation source data. Seamless DEM data layers under the 3DEP program  are available nationally at grid spacings of 1 arc-second (approximately 30 meters) for the conterminous United States, and at 1/3, 1/9 arc-seconds  (approximately 10 and 3 meters, respectively) and 1 meter for parts of the United States. Most seamless DEM data for Alaska is available at 2-arc-second (about 60 meters) grid spacing, where only lower resolution source data exist. Part of Alaska is available at the 1/3 arc-second, 1 arc-second and 5 meter resolution. Efforts are continuing to have full coverage of the 5 meter elevation data over Alaska in the next couple of years.
"""

USGS_PROVIDER = Provider(
    name="USGS",
    roles=["producer", "processor", "host"],
    url="https://www.usgs.gov/core-science-systems/ngp/3dep")

USGS_FTP_SERVER = "rockyftp.cr.usgs.gov"
USGS_FTP_BASE = f"ftp://{USGS_FTP_SERVER}/vdelivery/Datasets/Staged/Elevation"
AWS_BUCKET = "prd-tnm"
AWS_PREFIX = "StagedProducts/Elevation"
AWS_BASE = f"https://{AWS_BUCKET}.s3.amazonaws.com/{AWS_PREFIX}"

DEFAULT_BASE = AWS_BASE


import stactools.core

from stactools.threedep.metadata import Metadata

stactools.core.use_fsspec()


def register_plugin(registry):
    from stactools.threedep import commands
    registry.register_subcommand(commands.create_threedep_command)


__all__ = [Metadata]
__version__ = "0.2.0"
"""Library version"""



import datetime
from typing import Union, Optional
from xml.etree import ElementTree
from xml.etree.ElementTree import Element

from shapely.geometry import box, mapping
from pystac import Asset, MediaType, Link

from stactools.core import io
from stactools.core.io import ReadHrefModifier
from stactools.core.projection import reproject_geom
from stactools.threedep.constants import THREEDEP_CRS, THREEDEP_EPSG, DEFAULT_BASE
from stactools.threedep import utils


class Metadata:
    """3DEP file metadata."""
    @classmethod
    def from_href(cls,
                  href: str,
                  read_href_modifier: Optional[ReadHrefModifier] = None
                  ) -> "Metadata":
        """Creates a metadata from an href to the XML metadata file."""
        text = io.read_text(href, read_href_modifier)
        element_tree = ElementTree.fromstring(text)
        return cls(element_tree)

    @classmethod
    def from_product_and_id(cls,
                            product: str,
                            id: str,
                            base: str = None) -> "Metadata":
        """Creates a Metadata from a product and id."""
        if base is None:
            base = DEFAULT_BASE
        href = utils.path(product, id, extension="xml", base=base)
        return cls.from_href(href)

    def __init__(self, xml: Element):
        """Creates a new metadata object from XML metadata."""
        self.title = xml.findtext("./idinfo/citation/citeinfo/title")
        self.description = xml.findtext("./idinfo/descript/abstract")
        bounding = xml.find("./idinfo/spdom/bounding")
        self.minx = bounding.findtext("./westbc")
        self.miny = bounding.findtext("./southbc")
        self.maxx = bounding.findtext("./eastbc")
        self.maxy = bounding.findtext("./northbc")
        self.pubdate = xml.findtext("./idinfo/citation/citeinfo/pubdate")
        self.begdate = xml.findtext(
            "./idinfo/timeperd/timeinfo/rngdates/begdate")
        self.enddate = xml.findtext(
            "./idinfo/timeperd/timeinfo/rngdates/enddate")
        self.current = xml.findtext("./idinfo/timeperd/current")
        self.rowcount = xml.findtext("./spdoinfo/rastinfo/rowcount")
        self.colcount = xml.findtext("./spdoinfo/rastinfo/colcount")
        self.latres = xml.findtext("./spref/horizsys/geograph/latres")
        self.longres = xml.findtext("./spref/horizsys/geograph/longres")
        tiff_href = xml.findtext(
            "./distinfo/stdorder/digform/digtopt/onlinopt/computer/networka/networkr"
        )
        parts = tiff_href.split('/')[-4:]
        self.product = parts[0]
        self.id = parts[2]

    @property
    def stac_id(self) -> str:
        """Returns the STAC ID of this metadata.

        This is the id plus the product, e.g. if the filename of the tif is
        "USGS_1_n40w105.tif", then the STAC id is "n40w105-1".
        """
        return "{}-{}".format(self.id, self.product)

    @property
    def geometry(self) -> dict:
        """Returns this item's geometry in WGS84."""
        original_bbox = [
            float(self.minx),
            float(self.miny),
            float(self.maxx),
            float(self.maxy)
        ]
        return reproject_geom(THREEDEP_CRS, "EPSG:4326",
                              mapping(box(*original_bbox)))

    @property
    def datetime(self) -> Union["datetime.datetime", None]:
        """Returns the collection publication datetime."""
        if self.current == "publication date":
            return _format_date(self.pubdate)
        else:
            raise NotImplementedError

    @property
    def start_datetime(self) -> Union["datetime.datetime", None]:
        """Returns the start datetime for this record.

        This can be a while ago, since the national elevation dataset was
        originally derived from direct survey data.
        """
        return _format_date(self.begdate)

    @property
    def end_datetime(self) -> Union["datetime.datetime", None]:
        """Returns the end datetime for this record."""
        return _format_date(self.enddate, end_of_year=True)

    @property
    def gsd(self) -> float:
        """Returns the nominal ground sample distance from these metadata."""
        if self.product == "1":
            return 30
        elif self.product == "13":
            return 10
        else:
            raise NotImplementedError

    def data_asset(self, base: str = DEFAULT_BASE) -> Asset:
        """Returns the data asset (aka the tiff file)."""
        return Asset(href=self._asset_href_with_extension(base, "tif"),
                     title=self.title,
                     description=None,
                     media_type=MediaType.COG,
                     roles=["data"])

    def metadata_asset(self, base: str = DEFAULT_BASE) -> Asset:
        """Returns the data asset (aka the tiff file)."""
        return Asset(href=self._asset_href_with_extension(base, "xml"),
                     media_type=MediaType.XML,
                     roles=["metadata"])

    def thumbnail_asset(self, base: str = DEFAULT_BASE) -> Asset:
        """Returns the thumbnail asset."""
        return Asset(href=self._asset_href_with_extension(base, "jpg"),
                     media_type=MediaType.JPEG,
                     roles=["thumbnail"])

    def gpkg_asset(self, base: str = DEFAULT_BASE) -> Asset:
        """Returns the geopackage asset."""
        description = (
            "Spatially-referenced polygonal footprints of the source data used "
            "to assemble the DEM layer. The attributes of each source dataset, "
            "such as original spatial resolution, production method, and date "
            "entered into the standard DEM, are linked to these footprints.")
        return Asset(href=self._asset_href_with_extension(base,
                                                          "gpkg",
                                                          id_only=True),
                     media_type=MediaType.GEOPACKAGE,
                     roles=["metadata"],
                     description=description)

    def via_link(self, base: str = DEFAULT_BASE) -> Link:
        """Returns the via link for this file."""
        return Link("via", self._asset_href_with_extension(base, "xml"))

    @property
    def projection_extension_dict(self) -> dict:
        """Returns a dictionary of values to be applied to the projection extension."""
        shape = [int(self.rowcount), int(self.colcount)]
        transform = [
            float(self.longres),
            0.0,
            float(self.minx),
            0.0,
            -float(self.latres),
            float(self.maxy),
        ]
        return {
            "epsg": THREEDEP_EPSG,
            "shape": shape,
            "transform": transform,
        }

    @property
    def region(self) -> str:
        """Returns this objects 3dep "region".

        Region is defined as a 10x10 lat/lon box that nominally contains this item.
        E.g. for n41w106, the region would be n40w110. This is used mostly for
        creating subcatalogs for STACBrowser.
        """
        import math
        n_or_s = self.id[0]
        lat = float(self.id[1:3])
        if n_or_s == "s":
            lat = -lat
        lat = math.floor(lat / 10) * 10
        e_or_w = self.id[3]
        lon = float(self.id[4:])
        if e_or_w == "w":
            lon = -lon
        lon = math.floor(lon / 10) * 10
        return f"{n_or_s}{abs(lat)}{e_or_w}{abs(lon)}"

    def _asset_href_with_extension(self,
                                   base: str,
                                   extension: str,
                                   id_only: bool = False) -> str:
        if base is None:
            base = DEFAULT_BASE
        return utils.path(self.product,
                          self.id,
                          base=base,
                          extension=extension,
                          id_only=id_only)


def _format_date(
        date: str,
        end_of_year: bool = False) -> Union["datetime.datetime", None]:
    if len(date) == 4:
        year = int(date)
        if end_of_year:
            month = 12
            day = 31
        else:
            month = 1
            day = 1
        if year < 1800 or year > datetime.date.today().year:
            return None  # There's some bad metadata in the USGS records
        else:
            return datetime.datetime(year,
                                     month,
                                     day,
                                     0,
                                     0,
                                     0,
                                     tzinfo=datetime.timezone.utc)
    elif len(date) == 8:
        year = int(date[0:4])
        month = int(date[4:6])
        day = int(date[6:8])
        return datetime.datetime(year,
                                 month,
                                 day,
                                 0,
                                 0,
                                 0,
                                 tzinfo=datetime.timezone.utc)
    else:
        return None



import click
import os.path

from pystac import Catalog, Collection, Extent, CatalogType, MediaType
from pystac.extensions.item_assets import AssetDefinition, ItemAssetsExtension

from stactools.core.io import read_text
from stactools.threedep.constants import (PRODUCTS, DESCRIPTION, USGS_PROVIDER,
                                          USGS_FTP_BASE, USGS_3DEP_ID)
from stactools.threedep import utils, stac


def create_threedep_command(cli):
    """Creates the threedep command line utility."""
    @cli.group("threedep", short_help="Work with USGS 3DEP elevation data.")
    def threedep():
        pass

    @threedep.command(
        "create-catalog",
        short_help="Create a STAC catalog for existing USGS 3DEP data")
    @click.argument("destination")
    @click.option("-s",
                  "--source",
                  help="The href of a directory tree containing metadata",
                  default=None)
    @click.option("-i",
                  "--id",
                  multiple=True,
                  help="Ids to fetch. If not provided, will fetch all IDs.")
    @click.option("--quiet/--no-quiet", default=False)
    def create_catalog_command(destination, source, id, quiet):
        """Creates a relative published 3DEP catalog in DESTINATION.

        If SOURCE is not provided, will use the metadata in AWS. SOURCE is
        expected to be a directory tree mirroring the structure on USGS, so
        it is best created using `stac threedep download-metadata`.
        """
        base_ids = id  # not sure how to rename arguments in click
        collections = {}
        items = {}
        for product in PRODUCTS:
            items[product] = []
            if base_ids:
                ids = base_ids
            else:
                ids = utils.fetch_ids(product)
            for id in ids:
                item = stac.create_item_from_product_and_id(
                    product, id, source)
                items[product].append(item)
                if not quiet:
                    print(item.id)
            extent = Extent.from_items(items[product])
            if product == "1":
                title = "1 arc-second"
                description = "USGS 3DEP 1 arc-second DEMs"
            elif product == "13":
                title = "1/3 arc-second"
                description = "USGS 3DEP 1/3 arc-second DEMs"
            else:
                raise NotImplementedError
            collection = Collection(
                id=f"{USGS_3DEP_ID}-{product}",
                title=title,
                keywords=["USGS", "3DEP", "NED", "DEM", "elevation"],
                providers=[USGS_PROVIDER],
                description=description,
                extent=extent,
                license="PDDL-1.0")
            item_assets = ItemAssetsExtension.ext(collection,
                                                  add_if_missing=True)
            item_assets.item_assets = {
                "data":
                AssetDefinition({
                    "type": MediaType.COG,
                    "roles": ["data"],
                }),
                "metadata":
                AssetDefinition({
                    "type": MediaType.XML,
                    "roles": ["metadata"]
                }),
                "thumbnail":
                AssetDefinition({
                    "type": MediaType.JPEG,
                    "roles": ["thumbnail"]
                }),
                "gpkg":
                AssetDefinition({
                    "type": MediaType.GEOPACKAGE,
                    "roles": ["metadata"]
                })
            }
            collections[product] = collection
        catalog = Catalog(id=USGS_3DEP_ID,
                          description=DESCRIPTION,
                          title="USGS 3DEP DEMs",
                          catalog_type=CatalogType.RELATIVE_PUBLISHED)
        for product, collection in collections.items():
            catalog.add_child(collection)
            collection.add_items(items[product])
        catalog.generate_subcatalogs("${threedep:region}")
        catalog.normalize_hrefs(destination)
        catalog.save()
        catalog.validate()

    @threedep.command("download-metadata",
                      short_help="Download all metadata for USGS 3DEP data")
    @click.argument("destination")
    @click.option("-i",
                  "--id",
                  multiple=True,
                  help="Ids to fetch. If not provided, will fetch all IDs.")
    @click.option("--quiet/--no-quiet", default=False)
    def download_metadata_command(destination, id, quiet):
        """Creates a 3DEP collection in DESTINATION."""
        base_ids = id  # not sure how to rename arguments in click
        for product in PRODUCTS:
            if base_ids:
                ids = base_ids
            else:
                ids = utils.fetch_ids(product)
            for id in ids:
                path = utils.path(product,
                                  id,
                                  extension="xml",
                                  base=destination)
                if os.path.exists(path):
                    if not quiet:
                        print("{} exists, skipping download...".format(path))
                    continue
                os.makedirs(os.path.dirname(path), exist_ok=True)
                source_path = utils.path(product,
                                         id,
                                         extension="xml",
                                         base=USGS_FTP_BASE)
                if not quiet:
                    print("{} -> {}".format(source_path, path))
                text = read_text(source_path)
                with open(path, "w") as f:
                    f.write(text)

    @threedep.command(
        "fetch-ids",
        short_help="Fetch all product ids and print them to stdout")
    @click.argument("product")
    @click.option("--usgs-ftp/--no-usgs-ftp",
                  default=False,
                  help="Fetch from the USGS FTP instead of AWS.")
    def fetch_ids_command(product: str, usgs_ftp: bool):
        """Fetches product ids and prints them to stdout."""
        for id in utils.fetch_ids(product, use_usgs_ftp=usgs_ftp):
            print(id)



import os.path

import boto3
from ftplib import FTP

from stactools.threedep.constants import USGS_FTP_SERVER, AWS_BUCKET, AWS_PREFIX
from stactools.threedep import utils


def fetch_ids(product: str, use_usgs_ftp: bool = False) -> [str]:
    """Returns all ids for the given product."""
    if use_usgs_ftp:
        return _fetch_ids_from_usgs_ftp(product)
    else:
        return _fetch_ids_from_aws(product)


def _fetch_ids_from_usgs_ftp(product: str) -> [str]:
    directory = f"vdelivery/Datasets/Staged/Elevation/{product}/TIFF"
    ftp = FTP(USGS_FTP_SERVER)
    ftp.login()
    ids = [
        file_name for (file_name, _) in ftp.mlsd(directory)
        if not file_name.startswith(".")
    ]
    ftp.close()
    return ids


def _fetch_ids_from_aws(product: str) -> [str]:
    path = os.path.dirname(utils.path(product, ""))
    prefix = os.path.join(AWS_PREFIX, path)
    client = boto3.client("s3")
    paginator = client.get_paginator("list_objects_v2")
    page_iterator = paginator.paginate(Bucket=AWS_BUCKET, Prefix=prefix)
    filtered_iterator = page_iterator.search(
        "Contents[?ends_with(Key, `.xml`)].Key")
    # The main tif/xml files for each id are named simply like "USGS_1_n41w106.tif".
    # If there's been updates, older versions will have a datetime on the end, e.g.
    # "USGS_1_n41w106_20210330.tif" or something similar. By splitting on underscores,
    # we're hoping to catch only the "main" files.
    return [
        os.path.basename(os.path.dirname(key)) for key in filtered_iterator
        if len(os.path.basename(key).split("_")) == 3
    ]


def path(product: str,
         id: str,
         base: str = None,
         extension: str = None,
         id_only: bool = False) -> str:
    """Returns the subpath for this product and id.

    E.g. path("1", "n41w106") == "1/TIFF/n41w106/USGS_1_n41w106"
    E.g. path("1", "n41w106", extension="tif") == "1/TIFF/n41w106/USGS_1_n41w106.tif"
    E.g. path("1", "n41w106", base="/base/dir") == "/base/dir/1/TIFF/n41w106/USGS_1_n41w106"
    E.g. path("1", "n41w106", id_only=True) == "/base/dir/1/TIFF/n41w106/n41w106"
    """
    if id_only:
        path = "{}/TIFF/{}/{}".format(product, id, id)
    else:
        path = "{}/TIFF/{}/USGS_{}_{}".format(product, id, product, id)
    if extension:
        path = "{}.{}".format(path, extension)
    if base:
        path = os.path.join(base, path)
    return path



from typing import Optional

from pystac import Item
from pystac.extensions.projection import ProjectionExtension
from shapely.geometry import shape

from stactools.core.io import ReadHrefModifier
from stactools.threedep.metadata import Metadata
from stactools.threedep.constants import DEFAULT_BASE


def create_item(href: str,
                read_href_modifier: Optional[ReadHrefModifier] = None,
                base: str = DEFAULT_BASE) -> Item:
    """Creates a STAC item from an href to an XML metadata file."""
    metadata = Metadata.from_href(href, read_href_modifier)
    return create_item_from_metadata(metadata, base)


def create_item_from_product_and_id(product: str,
                                    id: str,
                                    base: str = DEFAULT_BASE) -> Item:
    """Creates a STAC item from a product (e.g. "1") and an ID (e.g. "n41w106")."""
    metadata = Metadata.from_product_and_id(product, id)
    return create_item_from_metadata(metadata, base)


def create_item_from_metadata(metadata: Metadata, base: DEFAULT_BASE) -> Item:
    """Creates a STAC item from Metadata."""
    geometry = metadata.geometry
    bbox = list(shape(geometry).bounds)
    item = Item(id=metadata.stac_id,
                geometry=geometry,
                bbox=bbox,
                datetime=metadata.datetime,
                properties={})
    start_datetime = metadata.start_datetime
    end_datetime = metadata.end_datetime
    if start_datetime and end_datetime:
        item.common_metadata.start_datetime = start_datetime
        item.common_metadata.end_datetime = end_datetime
    item.common_metadata.gsd = metadata.gsd
    item.links.append(metadata.via_link(base))
    item.assets["data"] = metadata.data_asset(base)
    item.assets["metadata"] = metadata.metadata_asset(base)
    item.assets["thumbnail"] = metadata.thumbnail_asset(base)
    item.assets["gpkg"] = metadata.gpkg_asset(base)
    projection = ProjectionExtension.ext(item, add_if_missing=True)
    projection.apply(**metadata.projection_extension_dict)
    item.properties["threedep:region"] = metadata.region
    return item
